# pylint: skip-file
from .base import BaseTestCase


class MySQLConnectionTestCase(BaseTestCase):

    def test_connect(self):
        self.add_connection('self', {
            'DB_ENGINE': 'mysql',
            'DB_HOST': 'localhost',
            'DB_NAME': 'rdbms',
            'DB_USERNAME': 'rdbms',
            'DB_PASSWORD': 'rdbms',
        })
        self.assertCanConnect('self')




"""Add relation books

Revision ID: bc5a0a045a3d
Revises: 
Create Date: 2021-11-17 01:22:22.915743

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'bc5a0a045a3d'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('books',
        sa.Column('id', sa.BigInteger(), nullable=False),
        sa.Column('title', sa.String(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('books')
    # ### end Alembic commands ###




# pylint: skip-file
import tempfile

from .base import BaseTestCase


class SQLiteConnectionTestCase(BaseTestCase):

    def test_connect_memory(self):
        self.add_connection('self', {
            'DB_ENGINE': 'sqlite',
            'DB_NAME': ':memory:'
        })
        self.assertCanConnect('self')

    def test_connect_disk(self):
        _, fn = tempfile.mkstemp()
        self.add_connection('self', {
            'DB_ENGINE': 'sqlite',
            'DB_NAME': fn
        })
        self.assertCanConnect('self')



# pylint: skip-file
import sqlalchemy
from unimatrix.ext.rdbms import declarative_base


Base, metadata = declarative_base()


class Book(Base):
    __tablename__ = 'books'

    id = sqlalchemy.Column(
        sqlalchemy.BigInteger,
        primary_key=True,
        name='id'
    )

    title = sqlalchemy.Column(
        sqlalchemy.String,
        name='title'
    )



# pylint: skip-file
import pytest

from .connectionmanager import connections
from .fixtures import setup_databases as rdbms
from .fixtures import rdbms_transaction


__all__ = [
    'rdbms',
    'rdbms_transaction',
]



"""Declares test fixtures to set up and tear down databases."""
import functools

import pytest
from unimatrix.ext import rdbms

from .testsession import TestSession


@pytest.mark.asyncio
@pytest.fixture(scope='function')
async def rdbms_transaction():
    methods = {}
    await rdbms.connect(debug=True)
    for alias, connection in dict.items(rdbms.connections.config):
        methods[alias] = connection.get_session
        session = connection.get_session()
        connection.get_session = TestSession(session)
    yield
    for alias, connection in dict.items(rdbms.connections.config):
        session = connection.get_session()
        await session.destroy()
        connection.get_session = methods[alias]

    await rdbms.disconnect()


@pytest.fixture(scope='package')
def setup_databases():
    config = rdbms.connections.config
    rdbms.connections.config = {
        alias: connection.as_test()
        for alias, connection in dict.items(config)
    }

    # Create the test databases
    for connection in dict.values(rdbms.connections.config):
        connection.create_database()
    yield
    for connection in dict.values(rdbms.connections.config):
        connection.drop_database()
    rdbms.connections.config = config



"""Declares :class:`Repository`."""
import sqlalchemy
from sqlalchemy.ext.asyncio import AsyncSession

from .connection import Connection
from .connectionmanager import connections


class Repository:
    """A repository implementation that uses a relational database as its
    storage backend.
    """
    db_alias: str = 'self'

    @property
    def session(self) -> AsyncSession:
        """Return the session that wraps the current transaction."""
        assert self.__session is not None # nosec
        return self.__session

    def new(self, **kwargs):
        """Return  a new :class:`Repository` instance."""
        return type(self)(
            *self.__initargs,
            **{**self.__initkwargs, **kwargs}
        )

    def __init__(self, session=None, *args, **kwargs):
        self.__session = session
        self.__transaction = None
        self.__initargs = args
        self.__initkwargs = kwargs
        self._setup(*args, **kwargs)

    def _setup(self, db_alias: str = 'self', *args, **kwargs):
        self.db_alias = db_alias
        self.setup(db_alias=db_alias, *args, **kwargs)

    def setup(self, db_alias: str = 'self', *args, **kwargs):
        """Hook that is called during instance initialization."""
        pass

    @property
    def connection(self) -> Connection:
        """Return the default database connection as specified by
        :attr:`db_alias`.
        """
        return connections.get(self.db_alias)

    def atomic(self):
        """Ensures that the statements executed within the context
        block are included in a single transaction.
        """
        return self.new(session=self.get_session())

    async def execute(self, query, *args, **kwargs):
        if isinstance(query, str):
            query = sqlalchemy.text(query)
        return await self.__session.execute(query, *args, **kwargs)

    def get_session(self, *args, **kwargs):
        """Return a :class:`sqlalchemy.ext.asyncio.AsyncSession`
        instance configured with the default database connection.
        """
        return self.connection.get_session(*args, **kwargs)

    async def __aenter__(self):
        assert self.__transaction is None # nosec
        self.__transaction = await self.__session.begin()
        return self

    async def __aexit__(self, cls, exception, traceback):
        if self.__transaction is not None:
            await self.__transaction.commit()\
                if exception is None else\
                await self.__transaction.rollback()
        await self.__session.close()



# pylint: skip-file
import tempfile

from .base import BaseTestCase


class ConfigurationTestCase(BaseTestCase):

    def test_configure_sqlite(self):
        self.add_connection('self', {
            'DB_ENGINE': 'sqlite',
            'DB_NAME': ''
        })

    def test_connect_sqlite_memory(self):
        self.add_connection('self', {
            'DB_ENGINE': 'sqlite',
            'DB_NAME': ':memory:'
        })
        self.assertCanConnect('self')

    def test_connect_sqlite_disk(self):
        _, fn = tempfile.mkstemp()
        self.add_connection('self', {
            'DB_ENGINE': 'sqlite',
            'DB_NAME': fn
        })
        self.assertCanConnect('self')

    def test_configure_postgresql(self):
        self.add_connection('self', {
            'DB_ENGINE': 'postgresql',
            'DB_HOST': 'localhost',
            'DB_NAME': 'rdbms',
            'DB_USERNAME': 'rdbms',
            'DB_PASSWORD': 'rdbms',
            'DB_PORT': "5432"
        })

    def test_configure_mysql(self):
        self.add_connection('self', {
            'DB_ENGINE': 'mysql',
            'DB_HOST': 'localhost',
            'DB_NAME': 'rdbms',
            'DB_USERNAME': 'rdbms',
            'DB_PASSWORD': 'rdbms',
            'DB_PORT': "3306"
        })



# pylint: skip-file
import asyncio

import sqlalchemy

from ..repository import Repository
from .base import BaseTestCase


class PostgreSQLTransactionIsolationTestCase(BaseTestCase):
    engine = 'postgresql'

    def setUp(self):
        super().setUp()
        self.add_connection('self', {
            'DB_ENGINE': self.engine,
            'DB_USERNAME': 'rdbms',
            'DB_PASSWORD': 'rdbms',
            'DB_NAME': 'rdbms'
        })
        self.connect_databases()
        self.connection = self.get_connection()
        self.relation = sqlalchemy.Table(
            'foo',
            self.metadata,
            sqlalchemy.Column(
                'id',
                sqlalchemy.String(7),
                primary_key=True
            )
        )
        self.run_async(self.setup)

    async def setup(self):
        async with self.connection.begin() as connection:
            await connection.run_sync(self.metadata.drop_all)
            await connection.run_sync(self.metadata.create_all)

    def test_mutation_is_not_visible(self):

        async def transaction_a(repo):
            async with repo.atomic() as session:
                await session.execute(
                    self.relation.insert()\
                        .values(id='foo')
                )
                await asyncio.sleep(0.5)
                result = list(await session.execute(self.relation.select()))
                assert result, result

        async def transaction_b(repo):
            async with repo.atomic() as session:
                result = list(await session.execute(self.relation.select()))
                assert not result, result
                await asyncio.sleep(1.5)
            async with repo.atomic() as session:
                result = list(await session.execute(self.relation.select()))
                assert result, result

        async def main():
            repo = Repository()
            await asyncio.gather(
                transaction_a(repo),
                transaction_b(repo),
            )

        self.run_async(main)


class MySQLTransactionIsolationTestCase(PostgreSQLTransactionIsolationTestCase):
    engine = 'mysql'



# pylint: skip-file
import asyncio
import os

import sqlalchemy
from unimatrix.ext import rdbms


async def tx1(db):
    async with db.atomic() as repo:
        await repo.execute("INSERT INTO foo VALUES (1)")
        await asyncio.sleep(1.5)


async def tx2(db):
    async with db.atomic() as repo:
        result = list(await repo.execute("SELECT * FROM foo"))
        assert not result, result


async def main():
    await rdbms.connect()
    async with rdbms.connections.get('self').begin() as connection:
        await connection.execute(sqlalchemy.text("TRUNCATE foo CASCADE"))

    repo = rdbms.Repository()
    await asyncio.gather(
        tx1(repo),
        tx2(repo)
    )


if __name__ == '__main__':
    asyncio.run(main())



"""Declares :class:`CommandLine` and the main script entrypoint."""
import os
from os.path import abspath
from os.path import dirname
from os.path import join

import ioc.loader
from alembic import config

from unimatrix.ext import rdbms


class CommandLine(config.CommandLine):

    def get_script_location(self, script_location):
        """Return the script location."""
        return script_location\
            or os.getenv('ALEMBIC_SCRIPT_LOCATION')\
            or abspath(dirname(__file__))

    def get_version_locations(self, cfg):
        """Get the ``version_locations`` configuration setting by
        discovering it from the metadata module.
        """
        module = ioc.loader.import_symbol(
            str.rsplit(cfg.get_main_option('metadata'), '.', 1)[0]
        )
        return abspath(join(dirname(module.__file__), 'migrations'))

    def parse_metadata(self, qualname: str):
        """Normalize the qualified name to the :class:`MetaData`
        objects.
        """
        if qualname.find(':') == -1:
            qualname = f"{qualname}:metadata"
        return str.replace(qualname, ':', '.')

    def main(self, argv=None):
        self.parser.add_argument('--metadata',
            help="The MetaData object(s) to create.",
            required=True,
            type=self.parse_metadata
        )
        self.parser.add_argument('--script-location',
            default=None,
            help="Override the script_location setting."
        )
        connection = rdbms.get('self')
        options = self.parser.parse_args(argv)
        if not hasattr(options, "cmd"):
            self.parser.error("too few arguments")
        else:
            cfg = config.Config()
            cfg.set_main_option(
                'script_location',
                self.get_script_location(options.script_location))
            cfg.set_main_option("sqlalchemy.url", connection.dsn)
            cfg.set_main_option("metadata", options.metadata)
            cfg.set_main_option(
                "version_locations",
                self.get_version_locations(cfg)
            )
            self.run_cmd(cfg, options)


def main(argv=None, prog=None, **kwargs):
    """The console runner function for Alembic."""
    CommandLine(prog=prog).main(argv=argv)


if __name__ == "__main__":
    main()



# pylint: skip-file
import unittest

import pytest

from unimatrix.ext import rdbms


@pytest.mark.usefixtures('rdbms')
class TestDatabaseSetupTestCase(unittest.TestCase):

    def setUp(self):
        self.connection = rdbms.get('self')
        self.db_name = self.connection.db_name

    def test_connections_are_replaced(self):
        self.assertTrue(str.startswith(self.db_name, 'test_'), self.db_name)

    def test_databases_are_created(self):
        pass



# pylint: skip-file
from sqlalchemy.ext.asyncio import AsyncSession

from .connectionmanager import connections
from .connection import Connection
from .connectionmanager import ConnectionManager
from .declarative import declarative_base
from .fixtures import setup_databases
from .repository import Repository


__all__ = [
    'Connection',
    'ConnectionManager',
    'Repository',
    'add',
    'connect',
    'declarative_base',
    'disconnect',
    'get',
    'session',
    'setup_databases',
]


def add(name: str, **opts) -> None:
    """Add a new connection using the given parameters"""
    connections.add(name, opts)


async def connect(*args, **kwargs) -> None:
    """Connect all database connections that are specified in the default
    connection manager.
    """
    return await connections.connect(*args, **kwargs)


async def disconnect() -> None:
    """Disconnect all database connections that are specified in the default
    connection manager.
    """
    return await connections.disconnect()


def get(name: str) -> Connection:
    """Return the named connection `name`."""
    return connections.get(name)


def session(self, name: str) -> AsyncSession:
    """Create a new :class:`sqlalchemy.ext.asyncio.AsyncSession`
    instance for the named connection `name`.
    """
    connection = get(name)
    if not connection.is_connected():
        raise RuntimeError(f"Connection '{name}' is not established.")
    return connection.get_session()



# pylint: skip-file
import asyncio
import threading

from .base import BaseTestCase



class ThreadingTestCase(BaseTestCase):

    def test_connection_is_local(self):
        self.add_connection('self', {
            'DB_ENGINE': 'sqlite',
            'DB_NAME': ':memory:'
        })
        event = threading.Event()
        done = threading.Event()
        connections = {}

        def thread_a():
            self.connect_databases()
            c = self.get_connection('self')
            connections[1] = c.is_connected()
            event.set()

        def thread_b():
            event.wait()
            c = self.get_connection('self')
            connections[2] = c.is_connected()

        t1 = threading.Thread(target=thread_a, daemon=True)
        t2 = threading.Thread(target=thread_b, daemon=True)
        t1.start()
        t2.start()
        t1.join()
        t2.join()

        self.assertTrue(connections[1])
        self.assertFalse(connections[2])
